#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <inttypes.h>
#include <stdio.h>
#include <assert.h>
#include <gcrypt.h>

#include "x2to1f.h"
#include "daps.h"

#define Hpub_FUN GCRY_MD_SHA256
#define Hpub_LEN (256 / 8)

#define Hhash_FUN GCRY_MD_SHA1
#define Hhash_LEN (160 / 8)

struct vk {
  struct pub *pub;
  size_t storagesizeA;
  size_t seedlenB;
  size_t siglen;
};

struct sk {
  struct td *td;
};

int KGen(struct vk **vk_p, struct sk **sk_p, int bits) {
  struct vk *vk;
  struct sk *sk;
  
  if (! (vk = malloc(sizeof(struct vk))))
    goto error_vk;
  if (! (sk = malloc(sizeof(struct sk))))
    goto error_sk;

  *vk_p = vk;
  *sk_p = sk;

  if (! TdGen(&vk->pub, &sk->td, bits))
    goto error_tdgen;

  vk->storagesizeA = get_storage_size_A(vk->pub);
  vk->seedlenB = get_sample_seed_len_B(vk->pub);
  vk->siglen = (8 * Hhash_LEN + 1) * vk->storagesizeA;

  return 1;

 error_tdgen:
  free(sk);
 error_sk:
  free(vk);
 error_vk:
  return 0;
}

void vk_release(struct vk *vk) {
  pub_release(vk->pub);
  free(vk);
}

void sk_release(struct sk *sk) {
  td_release(sk->td);
  free(sk);
}

size_t sig_length(const struct vk *vk) {
  return vk->siglen;
}

static void hash_string(gcry_md_hd_t hd, const char *s, size_t len) {
  assert(len <= 0xffFFffFF);
  gcry_md_putc(hd, (len >> 24) & 0xff);
  gcry_md_putc(hd, (len >> 16) & 0xff);
  gcry_md_putc(hd, (len >>  8) & 0xff);
  gcry_md_putc(hd, (len >>  0) & 0xff);
  gcry_md_write(hd, s, len);
}

static struct elemB* Hpub(const struct vk *vk, const gcry_md_hd_t hd) {
  gcry_md_hd_t hd2;
  char seed[vk->seedlenB];
  char *seedptr;
  size_t missing, amount;
  int ctr;

  assert(vk->seedlenB <= 256 * Hpub_LEN);

  seedptr = seed;
  missing = vk->seedlenB;
  for(ctr = 0; missing > 0; ctr++) {
    gcry_md_copy(&hd2, hd);
    gcry_md_putc(hd2, ctr);
    gcry_md_final(hd2);
    amount = (missing > Hpub_LEN) ? Hpub_LEN : missing;
    memcpy(seedptr, gcry_md_read(hd2, 0), amount);
    seedptr += amount;
    missing -= amount;
    gcry_md_close(hd2);
  }
  return SampleBFromSeed(vk->pub, seed);
}

static struct elemB* Hpub1(const struct vk *vk, const char *subj, size_t slen) {
  gcry_md_hd_t hd;
  struct elemB *b;
  gcry_md_open(&hd, Hpub_FUN, 0);
  hash_string(hd, subj, slen);
  b = Hpub(vk, hd);
  gcry_md_close(hd);
  return b;
}

static struct elemB* Hpub3(const struct vk *vk, const char *subj, size_t slen, const char *s, uint16_t i) {
  gcry_md_hd_t hd;
  struct elemB *b;
  gcry_md_open(&hd, Hpub_FUN, 0);
  hash_string(hd, subj, slen);
  gcry_md_write(hd, s, vk->storagesizeA);
  gcry_md_putc(hd, (i >> 8) & 0xff);
  gcry_md_putc(hd, (i >> 0) & 0xff);
  b = Hpub(vk, hd);
  gcry_md_close(hd);
  return b;
}

static void Hhash(const struct vk *vk, char *hash, const char *subj, size_t slen, const char *s, const char *msg, size_t mlen) {
  gcry_md_hd_t hd;
  gcry_md_open(&hd, Hhash_FUN, 0);
  hash_string(hd, subj, slen);
  gcry_md_write(hd, s, vk->storagesizeA);
  hash_string(hd, msg, mlen);
  gcry_md_final(hd);
  memcpy(hash, gcry_md_read(hd, 0), Hhash_LEN);
  gcry_md_close(hd);
}

void Sign(const struct sk *sk, const struct vk *vk, char *sig, const char *subj, size_t slen, const char *msg, size_t mlen) {
  struct elemA *a, *s;
  struct elemB *b;
  char hash[Hhash_LEN];
  uint16_t i;
  int d;

  b = Hpub1(vk, subj, slen);
  s = Reverse(sk->td, vk->pub, b, 0);
  elemA_export(vk->pub, s, sig);
  elemB_release(b);
  elemA_release(s);

  Hhash(vk, hash, subj, slen, sig, msg, mlen);

  for(i = 0; i < 8 * Hhash_LEN; i++) {
    d = (hash[i / 8] >> (i % 8)) & 1;
    b = Hpub3(vk, subj, slen, sig, i);
    a = Reverse(sk->td, vk->pub, b, d);
    elemA_export(vk->pub, a, sig + (i + 1) * vk->storagesizeA);
    elemB_release(b);
    elemA_release(a);
  }
}

int Ver(const struct vk *vk, const char *sig, const char *subj, size_t slen, const char *msg, size_t mlen) {
  struct elemA *a, *s;
  struct elemB *b1, *b2;
  char hash[Hhash_LEN];
  uint16_t i;
  int d, res;
  
  if (! (s = elemA_import(vk->pub, sig)))
    return 0;
  b1 = Apply(vk->pub, s);
  b2 = Hpub1(vk, subj, slen);
  res = elemB_equal(b1, b2);
  if (Decide(vk->pub, s))
    res = 0;
  elemA_release(s);
  elemB_release(b1);
  elemB_release(b2);
  
  if (! res)
    return 0;

  Hhash(vk, hash, subj, slen, sig, msg, mlen);

  for(i = 0; i < 8 * Hhash_LEN; i++) {
    d = (hash[i / 8] >> (i % 8)) & 1;

    if (! (a = elemA_import(vk->pub, sig + (i + 1) * vk->storagesizeA)))
      return 0;

    b1 = Apply(vk->pub, a);
    b2 = Hpub3(vk, subj, slen, sig, i);
    res = elemB_equal(b1, b2);

    if (Decide(vk->pub, a) != d)
      res = 0;
      
    elemA_release(a);
    elemB_release(b1);
    elemB_release(b2);

    if (! res)
      return 0;
  }
  return 1;
}

struct sk* skExtract(const struct vk *vk, const char *subj, size_t slen, const char *sig1, const char *msg1, size_t mlen1, const char *sig2, const char *msg2, size_t mlen2) {
  char hash1[Hhash_LEN], hash2[Hhash_LEN];
  struct elemA *a0, *a1;
  struct sk *sk;
  uint16_t i;
  int d1, d2;
  
  if ((mlen1 == mlen2) && ! memcmp(msg1, msg2, mlen1))
    return NULL;
  
  if (! Ver(vk, sig1, subj, slen, msg1, mlen1) || ! Ver(vk, sig2, subj, slen, msg2, mlen2))
    return NULL;

  if (! (sk = malloc(sizeof(struct sk))))
    return NULL;

  assert(! memcmp(sig1, sig2, vk->storagesizeA));

  Hhash(vk, hash1, subj, slen, sig1, msg1, mlen1);
  Hhash(vk, hash2, subj, slen, sig2, msg2, mlen2);
  assert(memcmp(hash1, hash2, Hhash_LEN));

  for(i = 0; ; i++) {
    d1 = (hash1[i / 8] >> (i % 8)) & 1;
    d2 = (hash2[i / 8] >> (i % 8)) & 1;
    if (d1 != d2)
      break;
  }

  if (d1 == 0) {
    a0 = elemA_import(vk->pub, sig1 + (i + 1) * vk->storagesizeA);
    a1 = elemA_import(vk->pub, sig2 + (i + 1) * vk->storagesizeA);
  } 
  else {
    a1 = elemA_import(vk->pub, sig1 + (i + 1) * vk->storagesizeA);
    a0 = elemA_import(vk->pub, sig2 + (i + 1) * vk->storagesizeA);
  }
  
  assert((sk->td = Extract(vk->pub, a0, a1)));

  elemA_release(a0);
  elemA_release(a1);
  
  return sk;
}
