#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <inttypes.h>
#include <stdio.h>
#include <assert.h>
#include <gcrypt.h>

#include "x2to1f.h"

struct pub {
  gcry_mpi_t n;
  gcry_mpi_t nm1d2;        /* (n-1)/2 */
  size_t storagesize;      /* exporting elements requires this much memory*/
  size_t sampleseedlen;    /* sampling routines for set A want this many bytes of entropy */
};

struct td {
  gcry_mpi_t p, q;
  gcry_mpi_t pp1d4, qp1d4; /* (p+1)/4 and (q+1)/4 */
};

struct elemA {
  gcry_mpi_t v;
};

struct elemB {
  gcry_mpi_t v;
};

static gcry_mpi_t mpi_import(const void *buf, size_t buflen) {
  gcry_mpi_t h;
  gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buf, buflen, NULL);
  unsigned int len = (gcry_mpi_get_nbits(h) + 7) / 8;
  assert(len <= buflen);
  assert(gcry_mpi_cmp_ui(h, 0) >= 0);
  return h;
}

static void mpi_export(void *buf, size_t buflen, const gcry_mpi_t x) {
  assert(gcry_mpi_cmp_ui(x, 0) >= 0);
  unsigned int len = (gcry_mpi_get_nbits(x) + 7) / 8;
  size_t NWRITTEN;
  assert(len <= buflen);
  memset(buf, 0, buflen);
  gcry_mpi_print(GCRYMPI_FMT_USG, buf + (buflen - len), len, &NWRITTEN, x);
  assert(NWRITTEN == len);
}

static void CRT_decompose(gcry_mpi_t *xp, gcry_mpi_t *xq, const gcry_mpi_t x, const gcry_mpi_t p, const gcry_mpi_t q) {
  *xp = gcry_mpi_new(0);
  *xq = gcry_mpi_new(0);
  gcry_mpi_mod(*xp, x, p);
  gcry_mpi_mod(*xq, x, q);
}

static void CRT_compose(gcry_mpi_t *x, const gcry_mpi_t xp, const gcry_mpi_t xq, const gcry_mpi_t p, const gcry_mpi_t q) {
  gcry_mpi_t a, u;
  a = gcry_mpi_new(0);
  u = gcry_mpi_new(0);
  *x = gcry_mpi_new(0);
  gcry_mpi_subm(a, xq, xp, q);
  gcry_mpi_invm(u, p, q);
  gcry_mpi_mulm(a, a, u, q); /* a = (xq - xp) / p  (mod q) */
  gcry_mpi_mul(*x, p, a);
  gcry_mpi_add(*x, *x, xp); /* x = p * ((xq - xp) / p mod q) + xp */
  gcry_mpi_release(a);
  gcry_mpi_release(u);
}

static int jacobi(gcry_mpi_t a, gcry_mpi_t n) {
  gcry_mpi_t x, y, h;
  int L = 1;
  x = gcry_mpi_new(0);
  y = gcry_mpi_new(0);
  h = gcry_mpi_new(0);
  gcry_mpi_set(x, a);
  gcry_mpi_set(y, n);
  for(;;) {
    gcry_mpi_rshift(h, y, 1);
    if (gcry_mpi_cmp(x, h) > 0) {
      gcry_mpi_sub(x, y, x);
      if (gcry_mpi_test_bit(y, 0) && gcry_mpi_test_bit(y, 1))
	L = -L;
    }
    while (! gcry_mpi_test_bit(x, 0) && ! gcry_mpi_test_bit(x, 1))
      gcry_mpi_rshift(x, x, 2);
    if (! gcry_mpi_test_bit(x, 0)) {
      gcry_mpi_rshift(x, x, 1);
      int yyy = 4 * gcry_mpi_test_bit(y, 2) + 2 * gcry_mpi_test_bit(y, 1) + 1 * gcry_mpi_test_bit(y, 0);
      if ((yyy == 3) || (yyy == 5))
	L = -L;
    }
    if (! gcry_mpi_cmp_ui(x, 1))
      break;
    if (gcry_mpi_test_bit(x, 0) && gcry_mpi_test_bit(x, 1) && gcry_mpi_test_bit(y, 0) && gcry_mpi_test_bit(y, 1))
      L = -L;
    gcry_mpi_mod(h, y, x);
    gcry_mpi_set(y, x);
    gcry_mpi_set(x, h);
  }
  gcry_mpi_release(x);
  gcry_mpi_release(y);
  gcry_mpi_release(h);
  return L;
}

/******************************************************************************/

int TdGen(struct pub **pub_p, struct td **td_p, int bits) {
  struct pub *pub;
  struct td *td;
  size_t buflen = bits / 16;
  unsigned char buf[buflen];

  assert(bits >= 256);
  assert(bits % 16 == 0);

  if (! (pub = malloc(sizeof(struct pub))))
    goto error_pub;
  if (! (td = malloc(sizeof(struct td))))
    goto error_td;

  *pub_p = pub;
  *td_p = td;

  gcry_randomize(buf, buflen, GCRY_STRONG_RANDOM);

  td->p = mpi_import(buf, buflen);
  gcry_mpi_set_highbit(td->p, bits / 2 - 1); /* set upper two bits, so that n=pq has maximum size */
  gcry_mpi_set_bit(td->p, bits / 2 - 2);
  gcry_mpi_clear_bit(td->p, bits / 2 - 3); /* clear third highest bit, so that p < q */
  gcry_mpi_clear_bit(td->p, 2); /* set lower bits such that result is 3 (mod 8) */
  gcry_mpi_set_bit(td->p, 1);
  gcry_mpi_set_bit(td->p, 0);
 
  while(gcry_prime_check(td->p, 0))
    gcry_mpi_add_ui(td->p, td->p, 8);

  gcry_randomize(buf, buflen, GCRY_STRONG_RANDOM);

  td->q = mpi_import(buf, buflen);
  gcry_mpi_set_highbit(td->q, bits / 2 - 1); /* set upper two bits, so that n=pq has maximum size */
  gcry_mpi_set_bit(td->q, bits / 2 - 2);
  gcry_mpi_set_bit(td->q, bits / 2 - 3); /* set third highest bit, so that p < q */
  gcry_mpi_set_bit(td->q, 2); /* set lower bits such that result is 7 (mod 8) */
  gcry_mpi_set_bit(td->q, 1);
  gcry_mpi_set_bit(td->q, 0);

  while(gcry_prime_check(td->q, 0))
    gcry_mpi_add_ui(td->q, td->q, 8);

  td->pp1d4 = gcry_mpi_new(0);
  gcry_mpi_add_ui(td->pp1d4, td->p, 1);
  gcry_mpi_rshift(td->pp1d4, td->pp1d4, 2);

  td->qp1d4 = gcry_mpi_new(0);
  gcry_mpi_add_ui(td->qp1d4, td->q, 1);
  gcry_mpi_rshift(td->qp1d4, td->qp1d4, 2);

  pub->n = gcry_mpi_new(0);
  gcry_mpi_mul(pub->n, td->p, td->q);

  pub->nm1d2 = gcry_mpi_new(0);
  gcry_mpi_sub_ui(pub->nm1d2, pub->n, 1);
  gcry_mpi_rshift(pub->nm1d2, pub->nm1d2, 1);

  pub->storagesize = bits / 8;
  pub->sampleseedlen = (bits + 128) / 8;

  return 1;

 error_td:
  free(pub);
 error_pub:
  return 0;
}

void pub_release(struct pub *pub) {
  gcry_mpi_release(pub->n);
  gcry_mpi_release(pub->nm1d2);
  free(pub);
}

void td_release(struct td *td) {
  gcry_mpi_release(td->p);
  gcry_mpi_release(td->q);
  gcry_mpi_release(td->pp1d4);
  gcry_mpi_release(td->qp1d4);
  free(td);
}

size_t get_sample_seed_len_A(const struct pub *pub) {
  return pub->sampleseedlen;
}

size_t get_sample_seed_len_B(const struct pub *pub) {
  return pub->sampleseedlen;
}

size_t get_storage_size_A(const struct pub *pub) {
  return pub->storagesize;
}

size_t get_storage_size_B(const struct pub *pub) {
  return pub->storagesize;
}

static gcry_mpi_t ZnSFromSeed(const struct pub *pub, const void *seed) {
  unsigned char iseed[pub->sampleseedlen];
  gcry_mpi_t a, h;
  if (! seed) {
    gcry_randomize(iseed, pub->sampleseedlen, GCRY_WEAK_RANDOM);
    seed = iseed;
  }
  a = mpi_import(seed, pub->sampleseedlen);
  gcry_mpi_mod(a, a, pub->n);
  assert(gcry_mpi_cmp_ui(a, 0));
  h = gcry_mpi_new(0);
  gcry_mpi_gcd(h, a, pub->n); /* gcrypt has a bug concerning the return value! */
  assert(! gcry_mpi_cmp_ui(h, 1));
  gcry_mpi_release(h);
  return a;
}

static gcry_mpi_t J1FromSeed(const struct pub *pub, const void *seed) {
  gcry_mpi_t a;
  a = ZnSFromSeed(pub, seed);
  if (jacobi(a, pub->n) < 0) {
    gcry_mpi_lshift(a, a, 1);
    if (gcry_mpi_cmp(a, pub->n) >= 0)
      gcry_mpi_sub(a, a, pub->n);
  }
  return a;
}

static gcry_mpi_t Jm1FromSeed(const struct pub *pub, const void *seed) {
  gcry_mpi_t a;
  a = ZnSFromSeed(pub, seed);
  if (jacobi(a, pub->n) > 0) {
    gcry_mpi_lshift(a, a, 1);
    if (gcry_mpi_cmp(a, pub->n) >= 0)
      gcry_mpi_sub(a, a, pub->n);
  }
  return a;
}

struct elemA* SampleAFromSeed(const struct pub *pub, const void *seed) {
  struct elemA *a;
  if (! (a = malloc(sizeof(struct elemA))))
    return NULL;
  a->v = ZnSFromSeed(pub, seed);
  if (gcry_mpi_cmp(a->v, pub->nm1d2) > 0)
    gcry_mpi_sub(a->v, pub->n, a->v);
  return a;
}

struct elemA* SampleA0FromSeed(const struct pub *pub, const void *seed) {
  struct elemA *a;
  if (! (a = malloc(sizeof(struct elemA))))
    return NULL;
  a->v = J1FromSeed(pub, seed);
  if (gcry_mpi_cmp(a->v, pub->nm1d2) > 0)
    gcry_mpi_sub(a->v, pub->n, a->v);
  return a;
}

struct elemA* SampleA1FromSeed(const struct pub *pub, const void *seed) {
  struct elemA *a;
  if (! (a = malloc(sizeof(struct elemA))))
    return NULL;
  a->v = Jm1FromSeed(pub, seed);
  if (gcry_mpi_cmp(a->v, pub->nm1d2) > 0)
    gcry_mpi_sub(a->v, pub->n, a->v);
  return a;
}

struct elemB* SampleBFromSeed(const struct pub *pub, const void *seed) {
  struct elemB *b;
  if (! (b = malloc(sizeof(struct elemB))))
    return NULL;
  b->v = J1FromSeed(pub, seed);
  if (gcry_mpi_cmp(b->v, pub->nm1d2) > 0)
    gcry_mpi_sub(b->v, pub->n, b->v);
  return b;
}

struct elemB* Apply(const struct pub *pub, const struct elemA *a) {
  struct elemB *b;
  if (! (b = malloc(sizeof(struct elemB))))
    return NULL;
  b->v = gcry_mpi_new(0);
  gcry_mpi_mulm(b->v, a->v, a->v, pub->n);
  if (gcry_mpi_cmp(b->v, pub->nm1d2) > 0)
    gcry_mpi_sub(b->v, pub->n, b->v);
  return b;
}

struct elemA* Reverse(const struct td *td, const struct pub *pub, const struct elemB *b, int d) {
  gcry_mpi_t bp, bq;
  struct elemA *a;

  if (! (a = malloc(sizeof(struct elemA))))
    return NULL;

  assert((d == 0) || (d == 1));
  assert(jacobi(b->v, pub->n) > 0);

  CRT_decompose(&bp, &bq, b->v, td->p, td->q);

  if (jacobi(bp, td->p) < 0) {
    gcry_mpi_sub(bp, td->p, bp);
    gcry_mpi_sub(bq, td->q, bq);
  }
  gcry_mpi_powm(bp, bp, td->pp1d4, td->p);
  gcry_mpi_powm(bq, bq, td->qp1d4, td->q);
  
  if ((jacobi(bp, td->p) == jacobi(bq, td->q)) == d)
    gcry_mpi_sub(bp, td->p, bp);

  CRT_compose(&a->v, bp, bq, td->p, td->q);

  if (gcry_mpi_cmp(a->v, pub->nm1d2) > 0)
    gcry_mpi_sub(a->v, pub->n, a->v);

  gcry_mpi_release(bp);
  gcry_mpi_release(bq);
  
  return a;
}

int Decide(const struct pub *pub, const struct elemA *a) {
  return (jacobi(a->v, pub->n) > 0) ? 0 : 1;
}

struct td* Extract(const struct pub *pub, const struct elemA *a0, const struct elemA *a1) {
  gcry_mpi_t h1, h2;
  struct td *td = NULL;

  h1 = gcry_mpi_new(0);
  h2 = gcry_mpi_new(0);

  if (! gcry_mpi_cmp(a0->v, a1->v))
    goto error;

  if (gcry_mpi_cmp(a0->v, a1->v) > 0)
    gcry_mpi_sub(h1, a0->v, a1->v);
  else
    gcry_mpi_sub(h1, a1->v, a0->v);
  gcry_mpi_gcd(h1, h1, pub->n);
  gcry_mpi_div(h2, NULL, pub->n, h1, 0);

  if (! gcry_mpi_cmp_ui(h1, 1))
    goto error;

  if (! (td = malloc(sizeof(struct td))))
    goto error;

  td->p = gcry_mpi_new(0);
  td->q = gcry_mpi_new(0);

  if (gcry_mpi_cmp(h1, h2) < 0) {
    gcry_mpi_set(td->p, h1);
    gcry_mpi_set(td->q, h2);
  }
  else {
    gcry_mpi_set(td->p, h2);
    gcry_mpi_set(td->q, h1);
  }

  td->pp1d4 = gcry_mpi_new(0);
  gcry_mpi_add_ui(td->pp1d4, td->p, 1);
  gcry_mpi_rshift(td->pp1d4, td->pp1d4, 2);

  td->qp1d4 = gcry_mpi_new(0);
  gcry_mpi_add_ui(td->qp1d4, td->q, 1);
  gcry_mpi_rshift(td->qp1d4, td->qp1d4, 2);

 error:
  gcry_mpi_release(h1);
  gcry_mpi_release(h2);
  return td;
}

void elemA_export(const struct pub *pub, const struct elemA *a, void *buf) {
  mpi_export(buf, pub->storagesize, a->v);
}

void elemB_export(const struct pub *pub, const struct elemB *b, void *buf) {
  mpi_export(buf, pub->storagesize, b->v);
}

struct elemA* elemA_import(const struct pub *pub, const void *buf) {
  struct elemA *a;
  gcry_mpi_t h;

  if (! (a = malloc(sizeof(struct elemA))))
    return NULL;
  a->v = mpi_import(buf, pub->storagesize);

  if (gcry_mpi_cmp_ui(a->v, 0) <= 0)
    goto error;
  if (gcry_mpi_cmp(a->v, pub->n) >= 0)
    goto error;
  
  h = gcry_mpi_new(0);
  gcry_mpi_gcd(h, a->v, pub->n); /* gcrypt has a bug concerning the return value! */
  if (gcry_mpi_cmp_ui(h, 1))
    goto error_h;
  gcry_mpi_release(h);

  return a;

 error_h:
  gcry_mpi_release(h);
 error:
  gcry_mpi_release(a->v);
  free(a);
  return NULL;
}

struct elemB* elemB_import(const struct pub *pub, const void *buf) {
  struct elemB *b;
  gcry_mpi_t h;

  if (! (b = malloc(sizeof(struct elemB))))
    return NULL;
  b->v = mpi_import(buf, pub->storagesize);

  if (gcry_mpi_cmp_ui(b->v, 0) <= 0)
    goto error;
  if (gcry_mpi_cmp(b->v, pub->n) >= 0)
    goto error;
  
  h = gcry_mpi_new(0);
  gcry_mpi_gcd(h, b->v, pub->n); /* gcrypt has a bug concerning the return value! */
  if (gcry_mpi_cmp_ui(h, 1))
    goto error_h;
  gcry_mpi_release(h);

  if (jacobi(b->v, pub->n) < 0)
    goto error;

  return b;

 error_h:
  gcry_mpi_release(h);
 error:
  gcry_mpi_release(b->v);
  free(b);
  return NULL;
}

int elemA_equal(const struct elemA *a1, const struct elemA *a2) {
  return ! gcry_mpi_cmp(a1->v, a2->v);
}

int elemB_equal(const struct elemB *b1, const struct elemB *b2) {
  return ! gcry_mpi_cmp(b1->v, b2->v);
}

void elemA_release(struct elemA *a) {
  gcry_mpi_release(a->v);
  free(a);
}

void elemB_release(struct elemB *b) {
  gcry_mpi_release(b->v);
  free(b);
}
